/**
 * <p>BoxFLuid demo combining 3D physics particles with the IsoSurface class to
 * create an animated mesh with a fluid behaviour. The mesh is optionally created
 * within a boundary sphere, but other forms can be created using a custom
 * ParticleConstraint class.</p>
 * 
 * <p>Dependencies:</p>
 * <ul>
 * <li>toxiclibscore-0015 or newer package from <a href="http://toxiclibs.org">toxiclibs.org</a></li>
 * <li>verletphysics-0004 or newer package from <a href="http://toxiclibs.org">toxiclibs.org</a></li>
 * <li>volumeutils-0002 or newer package from <a href="http://toxiclibs.org">toxiclibs.org</a></li>
 * <li>controlP5 GUI library from <a href="http://sojamo.de">sojamo.de</a></li>
 * </ul>
 * 
 * <p>Key controls:</p>
 * <ul>
 * <li>w : wireframe on/off</li>
 * <li>c : close sides on/off</li>
 * <li>p : show particles only on/off</li>
 * <li>b : turn bounding sphere on/off</li>
 * <li>r : reset particles</li>
 * <li>s : save current mesh as OBJ & STL format</li>
 * <li>- / = : decrease/increase surface threshold/tightness</li>
 * </ul>
 */

/* 
 * Copyright (c) 2009 Karsten Schmidt
 * 
 * This demo & library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * http://creativecommons.org/licenses/LGPL/2.1/
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

import processing.opengl.*;
import java.util.*;

import toxi.physics.*;
import toxi.physics.behaviors.*;
import toxi.physics.constraints.*;
import toxi.geom.*;
import toxi.geom.mesh.*;
import toxi.math.*;
import toxi.volume.*;
import toxi.processing.*;
import toxi.physics2d.*;

import controlP5.*;
import peasy.*;
import gab.opencv.*;
import KinectProjectorToolkit.*;
import SimpleOpenNI.*;


int NUM_PARTICLES = 100;
float REST_LENGTH=375;
int DIM=600;

//int GRID= 35;
int GRID = 10;
float VS=2*DIM/GRID;
Vec3D SCALE=new Vec3D(DIM, DIM, DIM).scale(2);
float isoThreshold= 6;
float alpha = 128;

int numP;
VerletPhysics physics;
ParticleConstraint boundingSphere;
GravityBehavior gravity;
AttractionBehavior keyAttractor, centerAttractor;

VolumetricSpaceArray volume;
IsoSurface surface;

TriangleMesh mesh=new TriangleMesh("fluid");

boolean showPhysics=false;
boolean isWireFrame=false;
boolean isClosed=true;
boolean useBoundary=false;

Vec3D colAmp=new Vec3D(400, 200, 200);


PeasyCam cam;
ControlP5 ui;
OpenCV opencv;
KinectProjectorToolkit kpc;
ToxiclibsSupport gfx;
SimpleOpenNI kinect;
Polygon2D body;


ArrayList<ProjectedContour> projectedContours;
ArrayList<VerletParticle2D> bodyPoints;
ArrayList<VerletSpring2D> bodySprings;
ArrayList<VerletParticle> bubblePoints;

boolean hideGUI, showFigure;
int mode;
color[] colors;

void setup() {
  size(displayWidth, displayHeight, OPENGL);
  smooth();
  kinect = new SimpleOpenNI(this);
  if (kinect.isInit() == false)
  {
    println("Can't init SimpleOpenNI, maybe the camera is not connected!"); 
    exit();
    return;
  } else {
    // mirror the image to be more intuitive
    kinect.setMirror(false);
    kinect.enableDepth();
    kinect.enableUser();
    kinect.alternativeViewPointDepthToImage();

    opencv = new OpenCV(this, kinect.depthWidth(), kinect.depthHeight());
    kpc = new KinectProjectorToolkit(this, kinect.depthWidth(), kinect.depthHeight());
    kpc.loadCalibration("calibration.txt");
    kpc.setContourSmoothness(2);

    gfx = new ToxiclibsSupport(this);
  }

  initPhysics();
  initGUI(); 
  volume=new VolumetricSpaceArray(SCALE, GRID, GRID, GRID);
  surface=new ArrayIsoSurface(volume);
  textFont(createFont("SansSerif", 12));

  cam = new PeasyCam(this, 1200);
  cam.setWheelScale(0.2);
  cam.rotateY(PI);
  bodyPoints = new ArrayList<VerletParticle2D>();
  bodySprings = new ArrayList<VerletSpring2D>();
  bubblePoints = new ArrayList<VerletParticle>();


  colors = new color[10];
  colors[0] = #0A86D0;
  colors[1] = #134ED5;
  colors[2] = #6D5CAF;
  colors[3] = #D2000D;
  colors[4] = #EB6002;
  colors[5] = #FFBC00;
  colors[6] = #019C69;
  colors[7] = #D42020;

  hideGUI = false;
  showFigure = false;
}

void draw() {
  background(230);

  computeVolume();
  updateContours();

  drawContours();
  updateAllBubbles();

  updateParticles();


  pushMatrix();
  //  translate(width*0.5, height*0.5, 0);

  //  rotateX(mouseY*0.01);
  //  rotateY(mouseX*0.01);
  noFill();
  stroke(255, 192);
  strokeWeight(1);
  //  box(physics.getWorldBounds().getExtent().x*2);
  if (showPhysics) {
    strokeWeight(4);
    stroke(50);
    for (Iterator i=physics.particles.iterator (); i.hasNext(); ) {
      VerletParticle p=(VerletParticle)i.next();
      Vec3D col=p.add(colAmp).scaleSelf(0.5);
      stroke(col.x, col.y, col.z);
      point(p.x, p.y, p.z);
    }
  } else {
    ambientLight(216, 216, 216);
    directionalLight(255, 255, 255, 0, 1, 0);
    directionalLight(96, 96, 96, 1, 1, -1);

    noStroke();
    fill(224, 0, 51);
    for (int i = 0; i< bubblePoints.size (); i++) {
      VerletParticle pos3d = bubblePoints.get(i);
      pushMatrix();
      translate(pos3d.x, pos3d.y, pos3d.z);
      color col = getColorByPosition(pos3d);
      fill(col, alpha);
      sphere(10);
      popMatrix();
    }

    if (isWireFrame) {
      stroke(255);
      noFill();
    } else {
      noStroke();
      fill(224, 0, 51);
    }


    beginShape(TRIANGLES);
    if (!isWireFrame) {
      drawFilledMesh();
    } else {
      drawWireMesh();
    }
    endShape();
  }



  popMatrix();
  noLights();
  fill(0);

  if (!hideGUI) {
    cam.beginHUD();
    text("faces: "+mesh.getNumFaces(), 20, 600);
    text("vertices: "+mesh.getNumVertices(), 20, 615);
    text("particles: "+physics.particles.size(), 20, 630);
    text("springs: "+physics.springs.size(), 20, 645);
    text("alpha: "+alpha, 20, 660);

    text("fps: "+frameRate, 20, 690);
    ui.draw();

    cam.endHUD();
  }

  bodyPoints.clear();
}

void keyPressed() {
  if (key=='r') initPhysics();
  if (key=='w') isWireFrame=!isWireFrame;
  if (key=='p') showPhysics=!showPhysics;
  if (key=='c') isClosed=!isClosed;
  if (key=='b') {
    toggleBoundary();
  }

  if (key == 'h') {
    hideGUI = !hideGUI;
  }
  if (key == 'f') {
    showFigure = !showFigure;
  }
  if (key == '1') {
    mode = 1;
  }
  if (key == '2') {
    mode = 2;
  }
  if (key == '3') {
    mode = 3;
  }
  if (key=='-' || key=='_') {
    alpha-=10;
  }
  if (key=='=' || key=='+') {
    alpha+=10;
  }

  if (key==',' || key=='<') {
    float strengh = centerAttractor.getStrength()-0.5;
    if (strengh>0.2)
      centerAttractor.setStrength(strengh);
  }
  if (key=='.' || key=='>') {
    float strengh = centerAttractor.getStrength()+0.5;
    if (strengh<5)
      centerAttractor.setStrength(strengh);
  }
  if (key=='s') {
    //    mesh.saveAsOBJ(sketchPath(mesh.name+".obj"));
    //    mesh.saveAsSTL(sketchPath(mesh.name+".stl"));
    saveFrame("######.jpg");
  }

  if (key == ' ') {
    keyAttractor = new AttractionBehavior(new Vec3D(0, 0, 0), 250, -10.f, 0.015);
    physics.addBehavior(keyAttractor);
  }
}

void keyReleased() {
  if (key == ' ') {
    physics.removeBehavior(keyAttractor);
  }
}

void toggleBoundary() {
  useBoundary=!useBoundary;
  initPhysics();
}

void updateContours() {
  kinect.update();  
  kpc.setDepthMapRealWorld(kinect.depthMapRealWorld()); 
  kpc.setKinectUserImage(kinect.userImage());
  opencv.loadImage(kpc.getImage());
  //  opencv.loadImage(createImage(640,480,RGB));
  // get projected contours
  projectedContours = new ArrayList<ProjectedContour>();
  ArrayList<Contour> contours = opencv.findContours();
  for (Contour contour : contours) {
    if (contour.area() > 2000) {
      ArrayList<PVector> cvContour = contour.getPoints();
      ProjectedContour projectedContour = kpc.getProjectedContour(cvContour, 1.0);
      projectedContours.add(projectedContour);
      //      println(contour.area());
    }
  }
}

void drawContours() {
  for (int i=0; i < projectedContours.size (); i++) {
    ProjectedContour projectedContour = projectedContours.get(i);
    int k = 0;
    float sum1=0, sum2=0;
    for (PVector p : projectedContour.getProjectedContours ()) {
      k++;
      //      println("k:"+k);
      sum1 = sum1 + p.x;
      sum2 = sum2 + p.y;
      if (k%2 == 0) {
        PVector t = projectedContour.getTextureCoordinate(p);
        stroke(0);
        strokeWeight(3);
        if ((random(1)<0.01) & (physics.particles.size()<NUM_PARTICLES)) {
          VerletParticle newBubble = new VerletParticle(p.x, p.y, 0);
          bubblePoints.add(newBubble);
          physics.addParticle(newBubble);
          //          physics.addBehavior(new AttractionBehavior(newBubble, 7, -0.2f, 0.05f));
        }
        bodyPoints.add(new VerletParticle2D(p.x, p.y));
      }
    }
    //    if (frameCount%5==0) {
    //      PVector p = projectedContour.getProjectedContours().get((frameCount*5)%k);
    //      //      println(frameCount%k+"   "+k);
    //      if (bubblePoints.size()<100) 
    //        addParticle(new Vec2D(p.x, p.y));
    //    }
  }

  body = null;
  body = new Polygon2D((List)bodyPoints);
  noStroke();
  fill(255);
  pushMatrix();
  translate(-width*0.5, -height*0.5, 0);
  if (showFigure)
    gfx.polygon2D(body);

  popMatrix();
}


void updateAllBubbles() {
  for (int i = 0; i< bubblePoints.size (); i++) {
    VerletParticle pos3d = bubblePoints.get(i);

    //    pushMatrix();
    //    translate(pos3d.x, pos3d.y, pos3d.z);
    //    sphere(10);
    //    popMatrix();

    VerletParticle2D pos = new VerletParticle2D(pos3d.x+width*0.5, pos3d.y+height*0.5);
    Vec3D prepos = pos3d.getPreviousPosition();
    VerletParticle2D pre = new VerletParticle2D(prepos.x+width*0.5, prepos.y+height*0.5);
    if ((body.containsPoint(pos) && !body.containsPoint(pre))) {
      //          pos.scaleVelocity(-2);
      pos3d.x = prepos.x;
      pos3d.y = prepos.y;
      pos3d.z = prepos.z;
    }
    if (!body.containsPoint(pos) && body.containsPoint(pre)) {
      Vec2D center = body.getCentroid();
      Vec3D velocity = new Vec3D(pos3d.x-center.x, pos3d.y-center.y, pos3d.z);
      pos3d.clearVelocity();
      pos3d.addVelocity(velocity.getNormalizedTo(2));
    }  
    if (body.containsPoint(pos) && body.containsPoint(pre)) {

      switch(mode) {
      case 1:
        pos3d.x = random(width);
        pos3d.y = random(height);
        break;
      case 2:
        float closestD = 1000;
        int pointIdx = 0;
        for (int ii = 0; ii <bodyPoints.size (); ii+=1) {
          VerletParticle2D tempBodyPoint = bodyPoints.get(ii);
          float distance = pos.distanceTo(new VerletParticle2D(tempBodyPoint.x - width*0.5, tempBodyPoint.y - height*0.5));
          if (distance < closestD) {
            closestD = distance;
            pointIdx = ii;
          }
        }
        Vec2D center = bodyPoints.get(pointIdx);
        //        pushMatrix();
        //        translate(center.x - width*0.5, center.y - height*0.5, 0);
        //        sphere(20);
        //        popMatrix();

        Vec3D velocity = new Vec3D(pos3d.x-center.x, pos3d.y-center.y, pos3d.z);
        pos3d.clearVelocity();
        pos3d.addVelocity(velocity.getNormalizedTo(-60));
        break;

      case 3:
        center = new Vec2D(0, 0);
        //        pushMatrix();
        //        translate(center.x - width*0.5, center.y - height*0.5, 0);
        //        sphere(20);
        //        popMatrix();

        velocity = new Vec3D(pos3d.x-center.x, pos3d.y-center.y, pos3d.z);
        pos3d.clearVelocity();
        pos3d.addVelocity(velocity.getNormalizedTo(60));
        break;
        //      pos.z = ran
      }

      pos3d.z = 0;
    }
  }
}

color getColorByPosition(VerletParticle pos3d) {

  color col;
  if (abs(pos3d.x) > 0.1) {
    float angle = (atan(pos3d.y/pos3d.x)/(PI/2)+1)/2*7;
    int clridx = int(angle)%7;
    println(clridx);
    col = lerpColor(colors[clridx], colors[(clridx+1)%7], angle - int(angle));
  } else {
    float angle = (atan(pos3d.y/0.1)/(PI/2)+1)/2*7;
    println(angle);
    int clridx = int(angle)%7;
    col = lerpColor(colors[clridx+1], colors[(clridx+1)%7+1], angle - int(angle));
  }
  return col;
}
